import re
from contextlib import contextmanager
from dataclasses import dataclass
from pathlib import Path, PosixPath

import pytest
from git import Actor, GitCommandError, GitError, Repo

from pygitops._constants import GIT_BRANCH_MAIN, GIT_BRANCH_MASTER
from pygitops._util import checkout_pull_branch, repo_working_dir
from pygitops.exceptions import PyGitOpsError, PyGitOpsStagedItemsError
from pygitops.operations import (
    feature_branch,
    get_default_branch,
    get_updated_repo,
    stage_commit_push_changes,
)

SOME_ACTOR = Actor("some-user", "some-user@company.com")
SOME_COMMIT_MESSAGE = "some-commit-message"
SOME_CONTENT_FILENAME = "foo.txt"
SOME_CHANGES = "some changes"
SOME_INITIAL_CONTENT = "foobar"
SOME_NEW_CONTENT = "newbar"
SOME_FEATURE_BRANCH = "test_branch"
SOME_DEFAULT_BRANCH_NAME = "some_default_branch_name"
SOME_SERVICE_ACCOUNT_NAME = "some-service-account-name"
SOME_SERVICE_ACCOUNT_TOKEN = "test_cred_123"

SOME_GITHUB_DOMAIN_NAME = "some-github-domain-name"
SOME_REPO_NAMESPACE = "test_namespace"
SOME_REPO_NAME = "some-repo-name"
SOME_CLONE_REPO_URL = f"https://{SOME_SERVICE_ACCOUNT_NAME}:{SOME_SERVICE_ACCOUNT_TOKEN}@{SOME_GITHUB_DOMAIN_NAME}/{SOME_REPO_NAMESPACE}/{SOME_REPO_NAME}.git"
SOME_CLONE_PATH = Path("foo")
SOME_LOCAL_REPO_NAME = "local"
SOME_HEAD_REF = f"refs/remotes/origin/{GIT_BRANCH_MASTER}"

SOME_OTHER_FILENAME = "SOME_OTHER_FILENAME.txt"
SOME_BRANCH_NAME = "some-branch-name"
SOME_OTHER_BRANCH_NAME = "some-other-branch-name"
SOME_CONTENT = "some-content"
SOME_MODIFY_FILE_DIFF = "diff --git a/foo.txt b/foo.txt\nindex 5f0c613..74cd6e7 100644\n--- a/foo.txt\n+++ b/foo.txt\n@@ -1 +1 @@\n-some changes\n\\ No newline at end of file\n+some-content\n\\ No newline at end of file"
SOME_DELETE_FILE_DIFF = "diff --git a/foo.txt b/foo.txt\ndeleted file mode 100644\nindex 5f0c613..0000000\n--- a/foo.txt\n+++ /dev/null\n@@ -1 +0,0 @@\n-some changes\n\\ No newline at end of file"
SOME_NEW_FILE_DIFF = "diff --git a/SOME_OTHER_FILENAME.txt b/SOME_OTHER_FILENAME.txt\nnew file mode 100644\nindex 0000000..e69de29"


@dataclass
class MultipleTestingRepos:
    """
    Model encapsulating the multi-repo setup used when testing git operations.

    :attr remote_repo: The bare remote repo.
    :attr local_repo: A local repo where state will be asserted.
    :attr cloned_repo: A cloned repo needed to push to bare remote to test _checkout_pull_branch.
    """

    remote_repo: Repo
    local_repo: Repo
    cloned_repo: Repo


def test_stage_commit_push_changes__push_failure__raises_pygitops_error(
    mocker, tmp_path
):
    mocker.patch("pygitops.operations._push_error_present", return_value=True)
    repos = _initialize_multiple_empty_repos(tmp_path)
    cloned_repo = repos.cloned_repo

    with feature_branch(cloned_repo, SOME_FEATURE_BRANCH):
        # Write to remote from the clone, then pull from local
        test_file_path = Path(repo_working_dir(cloned_repo)) / SOME_CONTENT_FILENAME
        test_file_path.write_text(SOME_INITIAL_CONTENT)

        with pytest.raises(PyGitOpsError, match=SOME_FEATURE_BRANCH):
            stage_commit_push_changes(
                cloned_repo, SOME_FEATURE_BRANCH, SOME_ACTOR, SOME_COMMIT_MESSAGE
            )


def _delete_existing_file(local_repo: Repo, filename: str) -> None:
    test_second_file_path = Path(repo_working_dir(local_repo)) / filename
    test_second_file_path.unlink()


def _modify_existing_file(local_repo: Repo, filename: str, content: str) -> None:
    test_file_path = Path(repo_working_dir(local_repo)) / filename
    test_file_path.write_text(content)


def _add_new_file(local_repo: Repo, filename: str) -> None:
    test_other_file_path = Path(repo_working_dir(local_repo)) / filename
    test_other_file_path.touch()


@pytest.mark.parametrize(
    (
        "some_source_modifier",
        "some_source_modifier_inputs",
        "some_other_source_modifier",
        "some_other_source_modifier_inputs",
        "some_result",
        "some_other_result",
    ),
    [
        (
            _modify_existing_file,
            [SOME_CONTENT_FILENAME, SOME_CONTENT],
            _modify_existing_file,
            [SOME_CONTENT_FILENAME, SOME_CONTENT],
            SOME_MODIFY_FILE_DIFF,
            SOME_MODIFY_FILE_DIFF,
        ),
        (
            _delete_existing_file,
            [SOME_CONTENT_FILENAME],
            _delete_existing_file,
            [SOME_CONTENT_FILENAME],
            SOME_DELETE_FILE_DIFF,
            SOME_DELETE_FILE_DIFF,
        ),
        (
            _add_new_file,
            [SOME_OTHER_FILENAME],
            _add_new_file,
            [SOME_OTHER_FILENAME],
            SOME_NEW_FILE_DIFF,
            SOME_NEW_FILE_DIFF,
        ),
        (
            _modify_existing_file,
            [SOME_CONTENT_FILENAME, SOME_CONTENT],
            _add_new_file,
            [SOME_OTHER_FILENAME],
            SOME_MODIFY_FILE_DIFF,
            SOME_NEW_FILE_DIFF,
        ),
        (
            _delete_existing_file,
            [SOME_CONTENT_FILENAME],
            _modify_existing_file,
            [SOME_CONTENT_FILENAME, SOME_CONTENT],
            SOME_DELETE_FILE_DIFF,
            SOME_MODIFY_FILE_DIFF,
        ),
        (
            _add_new_file,
            [SOME_OTHER_FILENAME],
            _delete_existing_file,
            [SOME_CONTENT_FILENAME],
            SOME_NEW_FILE_DIFF,
            SOME_DELETE_FILE_DIFF,
        ),
    ],
)
def test_feature_branch_context_manager__add_new_file_in_different_branches__clean_up_successfully(
    tmp_path,
    some_source_modifier,
    some_source_modifier_inputs,
    some_other_source_modifier,
    some_other_source_modifier_inputs,
    some_result,
    some_other_result,
):
    local_repo = _initialize_multiple_empty_repos(tmp_path).local_repo
    some_branch_name = SOME_BRANCH_NAME
    with feature_branch(local_repo, some_branch_name):
        some_source_modifier(local_repo, *some_source_modifier_inputs)
        some_diff = _get_diff(local_repo)
        assert some_diff == some_result
    some_other_branch_name = SOME_OTHER_BRANCH_NAME
    with feature_branch(local_repo, some_other_branch_name):
        some_other_source_modifier(local_repo, *some_other_source_modifier_inputs)
        some_other_diff = _get_diff(local_repo)
        assert some_other_diff == some_other_result


def test_stage_commit_push_changes__add_new_file__change_persisted(tmp_path):
    """
    Configure 'local' and 'remote' repositories with initial content.

    After generating new content in a local feature branch,
    verifies that `stage_commit_push_changes` function pushes new content to remote.
    """

    repos = _initialize_multiple_empty_repos(tmp_path)
    remote_repo = repos.remote_repo
    local_repo = repos.local_repo

    with feature_branch(local_repo, SOME_FEATURE_BRANCH):

        test_file_path = Path(repo_working_dir(local_repo)) / SOME_CONTENT_FILENAME
        test_file_path.write_text(SOME_INITIAL_CONTENT)

        # act, asserting expected state before and after operation
        with _assert_branch_state_stage_commit_push_changes__file_added(
            remote_repo, local_repo
        ):
            stage_commit_push_changes(
                local_repo, SOME_FEATURE_BRANCH, SOME_ACTOR, SOME_COMMIT_MESSAGE
            )

            # checkout `SOME_FEATURE_BRANCH` so that `_assert_branch_state_stage_commit_push_changes__file_added` compares correct commits
            remote_repo.heads[SOME_FEATURE_BRANCH].checkout()


def test_stage_commit_push_changes__remove_old_file__change_persisted(tmp_path):
    """
    Configure 'local' and 'remote' repositories with initial content.

    After generating new content in a local feature branch,
    verifies that `stage_commit_push_changes` function pushes new content to remote.
    """

    repos = _initialize_multiple_empty_repos(tmp_path)
    remote_repo = repos.remote_repo
    local_repo = repos.local_repo
    cloned_repo = repos.cloned_repo

    with feature_branch(cloned_repo, SOME_FEATURE_BRANCH):
        test_file_path = Path(repo_working_dir(cloned_repo)) / SOME_CONTENT_FILENAME
        test_file_path.write_text(SOME_INITIAL_CONTENT)

        stage_commit_push_changes(
            cloned_repo, SOME_FEATURE_BRANCH, SOME_ACTOR, SOME_COMMIT_MESSAGE
        )

        # Pull file on remote (pushed from clone) into local repo
        checkout_pull_branch(local_repo, SOME_FEATURE_BRANCH)

        # Assert expected state before and after operation
        with _assert_branch_state_stage_commit_push_changes__file_removed(
            remote_repo, local_repo
        ):
            removal_path = Path(repo_working_dir(local_repo)) / SOME_CONTENT_FILENAME
            removal_path.unlink()

            # Commit the removal
            stage_commit_push_changes(
                local_repo, SOME_FEATURE_BRANCH, SOME_ACTOR, SOME_COMMIT_MESSAGE
            )

            # checkout `SOME_FEATURE_BRANCH` so that `_assert_branch_state_stage_commit_push_changes__file_added` compares correct commits
            remote_repo.heads[SOME_FEATURE_BRANCH].checkout()


def test_stage_commit_push_changes__with_staged_files__adds_only_requested_file(
    tmp_path,
):
    """
    Configure 'local' and 'remote' repositories with initial content.

    After generating new content in a local feature branch,
    verifies that `stage_commit_push_changes` function pushes only the files specified to remote.
    """

    repos = _initialize_multiple_empty_repos(tmp_path)
    remote_repo = repos.remote_repo
    local_repo = repos.local_repo
    other_file = "my_test_file.txt"

    with feature_branch(local_repo, SOME_FEATURE_BRANCH):

        test_file_path = Path(repo_working_dir(local_repo)) / SOME_CONTENT_FILENAME
        test_file_path.write_text(SOME_INITIAL_CONTENT)
        other_file_path = Path(repo_working_dir(local_repo)) / other_file
        other_file_path.write_text("other test content")

        # act, asserting expected state before and after operation
        with _assert_branch_state_stage_commit_push_changes__file_added(
            remote_repo, local_repo
        ):
            stage_commit_push_changes(
                local_repo,
                SOME_FEATURE_BRANCH,
                SOME_ACTOR,
                SOME_COMMIT_MESSAGE,
                [test_file_path],
            )

            # checkout `SOME_FEATURE_BRANCH` so that `_assert_branch_state_stage_commit_push_changes__file_added` compares correct commits
            remote_repo.heads[SOME_FEATURE_BRANCH].checkout()

        # Assert the "other file" is still untracked
        assert [other_file] == local_repo.untracked_files


def test_stage_commit_push_changes__no_files_to_stage__raises_pygitops_error(tmp_path):

    repos = _initialize_multiple_empty_repos(tmp_path)
    local_repo = repos.local_repo

    with pytest.raises(PyGitOpsStagedItemsError):
        with feature_branch(local_repo, SOME_FEATURE_BRANCH):
            stage_commit_push_changes(
                local_repo, SOME_FEATURE_BRANCH, SOME_ACTOR, SOME_COMMIT_MESSAGE
            )


def test_stage_commit_push_changes__force_push_flag__changes_pushed(tmp_path):
    repos = _initialize_multiple_empty_repos(tmp_path)
    local_repo = repos.local_repo
    with feature_branch(local_repo, SOME_FEATURE_BRANCH):
        test_file_path = Path(repo_working_dir(local_repo)) / SOME_CONTENT_FILENAME
        test_file_path.write_text("content one")
        stage_commit_push_changes(
            local_repo, SOME_FEATURE_BRANCH, SOME_ACTOR, SOME_COMMIT_MESSAGE
        )

    branch = local_repo.create_head(SOME_FEATURE_BRANCH, force=True)
    branch.checkout()
    test_file_path.write_text("content two")
    stage_commit_push_changes(
        local_repo,
        SOME_FEATURE_BRANCH,
        SOME_ACTOR,
        SOME_COMMIT_MESSAGE,
        kwargs_to_push={"force": True},
    )


def test_feature_branch__untracked_files_present__raises_pygitops_error(mocker):
    untracked_file = "foo.py"
    repo = mocker.Mock(
        untracked_files=[untracked_file],
        git=mocker.Mock(symbolic_ref=mocker.Mock(return_value=SOME_HEAD_REF)),
    )

    with pytest.raises(PyGitOpsError, match=untracked_file):
        with feature_branch(repo, SOME_FEATURE_BRANCH):
            pass


def test_feature_branch__active_branch_not_master__raises_pygitops_error(mocker):

    active_branch_name = "some_active_feature_branch"
    active_branch = mocker.Mock()
    repo = mocker.Mock(
        untracked_files=[],
        active_branch=mocker.Mock(),
        heads={active_branch_name: active_branch, GIT_BRANCH_MASTER: None},
        git=mocker.Mock(symbolic_ref=mocker.Mock(return_value=SOME_HEAD_REF)),
    )

    with pytest.raises(PyGitOpsError):
        with feature_branch(repo, SOME_FEATURE_BRANCH):
            pass


def test_feature_branch__master_not_present_in_origin_refs__local_master_not_updated(
    mocker,
):

    origin_mock = mocker.Mock(refs=[])
    remotes_mock = mocker.Mock(origin=origin_mock)
    local_master_branch = mocker.Mock()

    repo = mocker.Mock(
        untracked_files=[],
        active_branch=local_master_branch,
        remotes=remotes_mock,
        working_dir=SOME_REPO_NAME,
        heads={GIT_BRANCH_MASTER: local_master_branch},
        git=mocker.Mock(symbolic_ref=mocker.Mock(return_value=SOME_HEAD_REF)),
    )

    _checkout_pull_branch_mock = mocker.patch(
        "pygitops.operations._checkout_pull_branch"
    )

    with feature_branch(repo, SOME_FEATURE_BRANCH):
        pass

    _checkout_pull_branch_mock.assert_not_called()


def test_feature_branch__master_present_in_origin_refs__local_master_updated(mocker):

    origin_mock = mocker.Mock(refs=[GIT_BRANCH_MASTER])
    remotes_mock = mocker.Mock(origin=origin_mock)
    local_master_branch = mocker.Mock()

    repo = mocker.Mock(
        untracked_files=[],
        active_branch=local_master_branch,
        remotes=remotes_mock,
        working_dir=SOME_REPO_NAME,
        heads={GIT_BRANCH_MASTER: local_master_branch},
        git=mocker.Mock(symbolic_ref=mocker.Mock(return_value=SOME_HEAD_REF)),
    )
    _checkout_pull_branch_mock = mocker.patch(
        "pygitops.operations._checkout_pull_branch"
    )

    with feature_branch(repo, SOME_FEATURE_BRANCH):
        pass

    _checkout_pull_branch_mock.assert_called_once_with(repo, GIT_BRANCH_MASTER)


@pytest.mark.parametrize(
    ["feature_branch_name", "checkout_expected"],
    ((SOME_FEATURE_BRANCH, True), (GIT_BRANCH_MASTER, False)),
)
def test_feature_branch__conditional_branch_checkout(
    mocker, feature_branch_name, checkout_expected
):

    origin_mock = mocker.Mock(refs=[GIT_BRANCH_MASTER])
    remotes_mock = mocker.Mock(origin=origin_mock)
    local_master_branch = mocker.Mock()
    feature_branch_mock = mocker.Mock()

    mocker.patch("pygitops.operations._checkout_pull_branch")
    repo = mocker.Mock(
        untracked_files=[],
        active_branch=local_master_branch,
        remotes=remotes_mock,
        working_dir=SOME_REPO_NAME,
        heads={GIT_BRANCH_MASTER: local_master_branch},
        create_head=mocker.Mock(return_value=feature_branch_mock),
        git=mocker.Mock(symbolic_ref=mocker.Mock(return_value=SOME_HEAD_REF)),
    )

    with feature_branch(repo, feature_branch_name):
        pass

    assert feature_branch_mock.checkout.called == checkout_expected


def test_feature_branch__exception_within_context__cleanup_occurs(mocker):

    some_branch_name = "some-feature-branch"
    origin_mock = mocker.Mock(refs=[GIT_BRANCH_MASTER])
    remotes_mock = mocker.Mock(origin=origin_mock)
    local_master_branch = mocker.Mock()

    mocker.patch("pygitops.operations._checkout_pull_branch")

    repo_mock = mocker.Mock(
        untracked_files=[],
        active_branch=local_master_branch,
        remotes=remotes_mock,
        working_dir=SOME_REPO_NAME,
        heads={GIT_BRANCH_MASTER: local_master_branch},
        create_head=mocker.Mock(),
        git=mocker.Mock(symbolic_ref=mocker.Mock(return_value=SOME_HEAD_REF)),
    )

    with pytest.raises(Exception):
        with feature_branch(repo_mock, some_branch_name):
            raise Exception("some exception")

    local_master_branch.checkout.assert_called_once()


def test_feature_branch__nested_calls__raises_pygitops_error(mocker):
    """Make sure the feature_branch context manager locks the repo correctly."""

    some_branch_name = "some-feature-branch"
    timeout_message = re.escape(
        f"The timeout of 10 seconds was exceeded when attempting to acquire the lockfile: lockfiles/{SOME_REPO_NAME}_lock_file.lock"
    )
    origin_mock = mocker.Mock(refs=[GIT_BRANCH_MASTER])
    remotes_mock = mocker.Mock(origin=origin_mock)
    local_master_branch = mocker.Mock()

    repo_mock = mocker.Mock(
        untracked_files=[],
        active_branch=local_master_branch,
        remotes=remotes_mock,
        working_dir=SOME_REPO_NAME,
        heads={GIT_BRANCH_MASTER: local_master_branch},
        git=mocker.Mock(symbolic_ref=mocker.Mock(return_value=SOME_HEAD_REF)),
    )

    with feature_branch(repo_mock, some_branch_name):
        with pytest.raises(
            PyGitOpsError,
            match=timeout_message,
        ):
            with feature_branch(repo_mock, some_branch_name):
                pass


def test_get_updated_repo__git_error_raised_by_repo__raises_pygitops_error(mocker):
    """GitError is the base exception class for the GitPython API so any exceptions raised by it will be GitErrors"""

    mocker.patch("pygitops.operations.Repo.clone_from", side_effect=GitError)

    with pytest.raises(PyGitOpsError):
        get_updated_repo(SOME_CLONE_REPO_URL, SOME_CLONE_PATH)


def test_get_updated_repo__repo_dne__fresh_clone_performed(mocker, tmp_path):

    clone_from_mock = mocker.patch("pygitops.operations.Repo.clone_from")

    get_updated_repo(SOME_CLONE_REPO_URL, tmp_path)

    clone_from_mock.assert_called_once_with(SOME_CLONE_REPO_URL, tmp_path)


def test_get_updated_repo__repo_dne__kwargs_passed_to_clone_from(mocker, tmp_path):

    clone_from_mock = mocker.patch("pygitops.operations.Repo.clone_from")

    some_kwargs = {
        "some_arg": "some-value",
        "another_arg": "another-value",
        "force": True,
    }

    get_updated_repo(SOME_CLONE_REPO_URL, tmp_path, **some_kwargs)

    clone_from_mock.assert_called_once_with(
        SOME_CLONE_REPO_URL,
        tmp_path,
        **{"some_arg": "some-value", "another_arg": "another-value"},
    )


def test_get_updated_repo__repo_exists_locally__repo_update_performed_against_default_branch(
    mocker, tmp_path
):

    Repo.init(tmp_path)
    master_branch_mock = mocker.Mock()
    repo_mock = mocker.Mock(
        heads={"master": master_branch_mock},
        git=mocker.Mock(symbolic_ref=mocker.Mock(return_value=SOME_HEAD_REF)),
    )
    mocker.patch("pygitops.operations.Repo", return_value=repo_mock)

    get_updated_repo(SOME_CLONE_REPO_URL, tmp_path)

    master_branch_mock.checkout.assert_called_once()
    repo_mock.remotes.origin.pull.assert_called_once()


@pytest.mark.parametrize(
    "force",
    [True, False],
)
def test_get_updated_repo__repo_exists_locally__repo_update_performed_against_provided_branch(
    mocker, tmp_path, force
):

    repo_mock = mocker.Mock()
    Repo.init(tmp_path)
    get_default_branch_mock = mocker.patch("pygitops.operations.get_default_branch")
    _checkout_pull_branch_mock = mocker.patch(
        "pygitops.operations._checkout_pull_branch"
    )
    mocker.patch("pygitops.operations.Repo", return_value=repo_mock)

    get_updated_repo(
        SOME_CLONE_REPO_URL, tmp_path, branch=SOME_FEATURE_BRANCH, force=force
    )

    get_default_branch_mock.assert_not_called()
    _checkout_pull_branch_mock.assert_called_once_with(
        repo_mock, SOME_FEATURE_BRANCH, force=force
    )


def test_get_updated_repo__clone_dirs_dne__clone_dirs_created(mocker, tmp_path):

    clone_dir = tmp_path / "some_parent" / "repo"
    repo_mock = mocker.Mock()
    mocker.patch("pygitops.operations.Repo", return_value=repo_mock)

    get_updated_repo(SOME_CLONE_REPO_URL, clone_dir)

    assert clone_dir.exists()


def test_get_updated_repo__file_operations__repo_not_present(tmp_path):
    """
    A fresh clone of a repository should pull down remote content

    Configures 'remote' repository with initial content
    Cloning the remote repo should create the expected content in the new local repository
    """

    # initialize remote repo
    remote_path = tmp_path / "remote"
    local_path = tmp_path / "local"
    local_path.mkdir()
    remote_repo = _initialize_repo_with_content(remote_path)

    # write and commit some content in the remote repo
    _commit_content(remote_repo, SOME_INITIAL_CONTENT)

    # clone remote repo to local
    local_repo = get_updated_repo(str(remote_path), local_path)

    filepath = f"{local_repo.working_tree_dir}/{SOME_CONTENT_FILENAME}"

    # content from remote should be present in local repo
    with open(filepath, "r") as test_file:
        content = test_file.read()
        assert SOME_INITIAL_CONTENT in content


def test_get_updated_repo__local_repo_on_disk__remote_default_branch_changes__remote_default_pulled(
    tmp_path,
):

    remote_path = tmp_path / "remote"
    local_path = tmp_path / "local"
    local_path.mkdir()
    remote_repo = _initialize_repo_with_content(remote_path)

    local_repo = get_updated_repo(str(remote_path), local_path)

    assert local_repo.active_branch.name != SOME_DEFAULT_BRANCH_NAME

    # create and update head ref on remote repo
    default_head = remote_repo.create_head(SOME_DEFAULT_BRANCH_NAME)
    remote_repo.head.set_reference(default_head)

    local_repo = get_updated_repo(str(remote_path), local_path)
    assert local_repo.active_branch.name == SOME_DEFAULT_BRANCH_NAME


def test_get_updated_repo__file_operations__repo_present_locally(tmp_path):
    """
    A clone of a repository that already exists locally should checkout master branch and pull down latest content

    Configures "remote" repository with initial content, and clones repo to local machine

    We will move the local repo to another branch, and make an additional commit on the master branch of the remote repo
    The side effect of running the clone operation is that the master branch is checked out and updated locally
    """

    # initialize remote repo
    remote_path = tmp_path / "remote"
    local_path = tmp_path / "local"
    local_path.mkdir()
    remote_repo = _initialize_repo_with_content(remote_path)

    # write and commit some content in the remote repo
    _commit_content(remote_repo, SOME_INITIAL_CONTENT)

    # setup the preconditions outlined in our test case description
    local_repo = get_updated_repo(str(remote_path), local_path)
    local_repo.create_head(SOME_FEATURE_BRANCH)
    local_repo.heads[SOME_FEATURE_BRANCH].checkout()
    _commit_content(remote_repo, SOME_NEW_CONTENT)

    # act, asserting expected state before and after operation
    with _assert_get_updated_repo_state(local_repo):
        get_updated_repo(str(remote_path), local_path)


def test_get_updated_repo__error__login_not_in_error(mocker):

    mocker.patch(
        "pygitops.operations.Repo.clone_from",
        side_effect=GitCommandError("some-command", "some-status"),
    )

    with pytest.raises(PyGitOpsError) as exc_info:
        get_updated_repo(SOME_CLONE_REPO_URL, SOME_CLONE_PATH)

    assert SOME_SERVICE_ACCOUNT_TOKEN not in str(exc_info.value)


def test_get_updated_repo__error__login_scrubbed(mocker):
    mocker.patch(
        "pygitops.operations.Repo.clone_from",
        side_effect=GitError(SOME_CLONE_REPO_URL),
    )

    with pytest.raises(PyGitOpsError) as exc_info:
        get_updated_repo(SOME_CLONE_REPO_URL, SOME_CLONE_PATH)

    exception_text = str(exc_info.value)
    assert "https://***:***@" in exception_text
    assert SOME_SERVICE_ACCOUNT_TOKEN not in exception_text


def test_get_updated_repo__clone_dir_as_str(mocker, tmp_path):

    clone_from_mock = mocker.patch(
        "pygitops.operations.Repo.clone_from",
    )

    mocker.patch("pygitops.operations._is_git_repo", mocker.Mock(return_value=False))
    get_updated_repo(SOME_CLONE_REPO_URL, str(tmp_path))

    assert clone_from_mock.called
    assert clone_from_mock.call_args[0][1] == PosixPath(str(tmp_path))


def test_get_default_branch__match_not_present__raises_pygitops_error(mocker):

    repo_mock = mocker.Mock(
        git=mocker.Mock(
            symbolic_ref=mocker.Mock(return_value="some-unmatched-symbolic-ref")
        )
    )

    with pytest.raises(PyGitOpsError):
        get_default_branch(repo_mock)


def test_get_default_branch__match_index_error__raises_pygitops_error(mocker):

    repo_mock = mocker.Mock(
        git=mocker.Mock(symbolic_ref=mocker.Mock(return_value="refs/remotes/origin/"))
    )

    mocker.patch(
        "pygitops.operations.re",
        new=mocker.Mock(
            match=mocker.Mock(
                return_value=mocker.Mock(group=mocker.Mock(side_effect=IndexError))
            )
        ),
    )

    with pytest.raises(PyGitOpsError):
        get_default_branch(repo_mock)


@pytest.mark.parametrize(
    "branch_name",
    ["main", "master", "2.x", "name-with-dashes", "name_with_underscores"],
)
def test_get_default_branch__default_branch_returned(tmp_path, branch_name):

    repos = _initialize_multiple_empty_repos(tmp_path, initial_branch=branch_name)
    remote_repo = repos.remote_repo
    local_repo = repos.local_repo

    # the local repo's default branch should reflect the HEAD ref on the remote branch
    assert get_default_branch(local_repo) == remote_repo.heads[0].name


def test_get_default_branch__remote_head_changes__new_default_branch_returned(tmp_path):

    repos = _initialize_multiple_empty_repos(tmp_path)
    remote_repo = repos.remote_repo
    local_repo = repos.local_repo

    # initially, the local repo's default branch should reflect the HEAD ref on the remote branch
    assert get_default_branch(local_repo) == remote_repo.heads[0].name

    # point remote repo's HEAD ref at new branch
    new_branch = remote_repo.create_head(GIT_BRANCH_MAIN)
    remote_repo.head.reference = new_branch

    assert get_default_branch(local_repo) == GIT_BRANCH_MAIN


def _initialize_repo_with_content(repo_path):
    """
    Helper function used to initialize repo objects, write content, and make initial commit
    """

    repo_path.mkdir()
    repo = Repo.init(repo_path, initial_branch=GIT_BRANCH_MASTER)

    _commit_content(repo, SOME_INITIAL_CONTENT, commit_message="Initial commit")

    return repo


def _commit_content(repo, content, commit_message="another commit"):
    """
    Helper function used to commit content to repository
    """
    with open(f"{repo.working_tree_dir}/{SOME_CONTENT_FILENAME}", "a+") as test_file:
        test_file.write(content)
    index = repo.index
    index.add([SOME_CONTENT_FILENAME])
    index.commit(commit_message)


@contextmanager
def _assert_get_updated_repo_state(repo):
    """
    Runs assertions before and after clone operation, in case where repo already exists locally
    """

    filepath = f"{repo.working_tree_dir}/{SOME_CONTENT_FILENAME}"

    # prior to running the clone operation, `SOME_FEATURE_BRANCH` is the active feature branch
    assert repo.active_branch == repo.heads[SOME_FEATURE_BRANCH]
    with open(filepath, "r") as test_file:
        content = test_file.read()
        assert SOME_INITIAL_CONTENT in content
        assert SOME_NEW_CONTENT not in content

    yield

    # after running the clone operation, `GIT_BRANCH_MASTER` is the active feature branch, and the new content is present
    assert repo.active_branch == repo.heads[GIT_BRANCH_MASTER]
    with open(filepath, "r") as test_file:
        content = test_file.read()
        assert SOME_INITIAL_CONTENT in content
        assert SOME_NEW_CONTENT in content


@contextmanager
def _assert_branch_state_stage_commit_push_changes__file_added(remote_repo, local_repo):
    """Run assertions before and after staging, committing, and pushing for a given repo pair."""

    local_content_path = Path(repo_working_dir(local_repo)) / SOME_CONTENT_FILENAME

    # The file should exist only on local before pushing
    assert SOME_INITIAL_CONTENT in local_content_path.read_text()

    yield

    # The remote repo should now have at least one commit,
    # and the commit history should be present on both local and remote after pushing

    assert len(list(remote_repo.iter_commits())) != 0
    assert list(remote_repo.iter_commits()) == list(local_repo.iter_commits())


@contextmanager
def _assert_branch_state_stage_commit_push_changes__file_removed(
    remote_repo, local_repo
):
    """Run assertions before and after staging, committing, and pushing for a given repo pair."""
    local_content_path = Path(repo_working_dir(local_repo)) / SOME_CONTENT_FILENAME

    # The file should exist only on local before pushing
    assert SOME_INITIAL_CONTENT in local_content_path.read_text()

    yield

    # The file should now be removed on local
    assert not local_content_path.exists()
    # The commit history should be the same on both local and remote after pushing
    # (including the commit to remove the file)
    assert list(remote_repo.iter_commits()) == list(local_repo.iter_commits())


# Helper functions for testing git operations
def _initialize_multiple_empty_repos(
    base_path, initial_branch="main"
) -> MultipleTestingRepos:
    """
    Helper function used to initialize and configure local, remote, and clone repos for integration testing.
    """

    # Setup Test Case
    local_path = base_path / "local"
    local_path.mkdir()

    remote_path = base_path / "remote.git"
    remote_path.mkdir()

    # We have to set up a clone because we can't commit directly to a bare repo
    clone_path = base_path / "clone"
    clone_path.mkdir()

    # The remote is set up as a bare repository
    remote_repo = Repo.init(remote_path, initial_branch=initial_branch)

    # give the remote repo some initial commit history
    new_file_path = remote_path / SOME_CONTENT_FILENAME
    new_file_path.write_text(SOME_CHANGES)
    remote_repo.index.add([str(new_file_path)])
    remote_repo.index.commit(
        SOME_COMMIT_MESSAGE, author=SOME_ACTOR, committer=SOME_ACTOR
    )

    # Simplify things by assuring local copies are cloned from the remote.
    # Because they are clones, they will have 'origin' correctly configured
    local_repo = Repo.clone_from(repo_working_dir(remote_repo), local_path)
    cloned_repo = Repo.clone_from(repo_working_dir(remote_repo), clone_path)

    return MultipleTestingRepos(
        remote_repo=remote_repo, local_repo=local_repo, cloned_repo=cloned_repo
    )


def _get_diff(repo: Repo) -> str:
    """
    Helper function used to handle the logic of generating diff text via a feature branch.

    This diff will only reflect the changes since the last commit.
    This includes staging the local changes.
    """

    index = repo.index
    workdir_path = Path(repo_working_dir(repo))

    untracked_file_paths = [Path(f) for f in repo.untracked_files]
    items_to_stage = untracked_file_paths + [Path(f.a_path) for f in index.diff(None)]

    for item in items_to_stage:
        full_path = workdir_path / item
        index.add(str(item)) if full_path.exists() else index.remove(str(item), r=True)

    return repo.git.diff("--cached")



import logging
import os
from contextlib import contextmanager
from os import PathLike
from pathlib import Path
from typing import Iterator, Union

from filelock import FileLock, Timeout
from git import PushInfo, Repo
from git.exc import InvalidGitRepositoryError

from pygitops.exceptions import PyGitOpsError, PyGitOpsWorkingDirError

_logger = logging.getLogger(__name__)
_lockfile_path = Path("lockfiles")

FILELOCK_ACQUIRE_TIMEOUT_SECONDS = 10


@contextmanager
def lock_repo(repo: Repo) -> Iterator[None]:
    """
    Lock a given repo for use.

    :param repo: The repo to lock on.
    """

    repo_name = os.path.basename(os.path.normpath(repo_working_dir(repo)))
    lockfile_name = str(get_lockfile_path(repo_name))

    lock = FileLock(lockfile_name)
    try:
        with lock.acquire(timeout=FILELOCK_ACQUIRE_TIMEOUT_SECONDS):
            _logger.debug(
                f"Successfully acquired lock: {lockfile_name} for repo: {repo}"
            )
            # yield control to the function using this context manager
            yield
            _logger.debug(f"About to release lock: {lockfile_name} for repo: {repo}")
    except Timeout:
        raise PyGitOpsError(
            f"The timeout of {FILELOCK_ACQUIRE_TIMEOUT_SECONDS} seconds was exceeded when attempting to acquire the lockfile: {lockfile_name}"
        )


def checkout_pull_branch(repo: Repo, branch: str, force: bool = False) -> None:
    """
    Pull changes from the specified branch of a repo.

    Will fail if the branch does not exist on the remote
    """

    origin = repo.remotes.origin

    # `origin.refs` might be out of date, this makes local checkout of repo aware of remote branches
    origin.fetch()

    if branch not in repo.heads:

        # handle case where provided branch name isnt a known remote branch
        if branch not in origin.refs:
            raise PyGitOpsError(
                f"The provided branch {branch} does not exist on the remote origin"
            )

        repo.create_head(branch, origin.refs[branch])
        _logger.debug(
            f"[Pull Branch] Create head was successful for repo: {repo}, branch: {branch}"
        )

        repo.heads[branch].set_tracking_branch(origin.refs[branch])
        _logger.debug(
            f"[Pull Branch] Set tracking branch successful for repo: {repo}, branch: {branch}"
        )

    # checkout local `branch` to working tree, optionally discarding changes to the index and working tree
    repo.heads[branch].checkout(force=force)
    _logger.debug(
        f"[Pull Branch] Checkout local branch successful for repo: {repo}, branch: {branch}"
    )

    if force:
        repo.git.clean("-df")
        _logger.debug(
            f"[Pull Branch] Removed untracked files for repo: {repo}, branch: {branch}"
        )

    # pull the changes from the remote branch
    origin.pull(branch)
    _logger.debug(
        f"[Pull Branch] Pull of changes successful for repo: {repo}, branch: {branch}"
    )


def get_lockfile_path(repo_name: str) -> Path:
    """Get a lockfile to lock a git repo."""

    _lockfile_path.mkdir(exist_ok=True)

    return _lockfile_path / f"{repo_name}_lock_file.lock"


def push_error_present(push_info: PushInfo) -> bool:
    """
    Given an instance of `git.remote.PushInfo`, determine if error is present.

    GitPython's remote operations use bitflags to indicate status of operations:
        https://gitpython.readthedocs.io/en/stable/reference.html#git.remote.PushInfo

    Check for presence of the error flag in the returned flags bitmask.
    """
    return bool(push_info.flags & push_info.ERROR)


def is_git_repo(path: Path) -> bool:
    """
    Determine if a given path is a valid git repository.

    :param path: Directory to inspect
    :return: True if the contents of a directory at given path contains a valid git repository
    """
    try:
        Repo(path).git_dir
        return True
    except InvalidGitRepositoryError:
        return False


def repo_working_dir(repo: Repo) -> Union[str, PathLike]:
    """
    Between gitpython 3.1.18 and 3.1.29, `git.Repo.working_dir` is now typed as Optional.

    The `os.path` and `pathlib.Path` operations are not typed to support `Optional`.
    Calling this function to access this property will handle cases where `repo.working_dir` is None, although it should never happen in practice.

    :param repo: The repo whose `working_dir` we are interested in.
    :raises PyGitOpsWorkingDirError: Raise error if `working_dir` is unexpectedly None.
    :return str: `working_dir` of the repo
    """
    working_dir = repo.working_dir

    # Can replace with assignment operator when python 3.7 support is dropped
    if working_dir is None:
        raise PyGitOpsWorkingDirError(
            f"The working_dir for repo {repo} is unexpectedly None"
        )

    return working_dir



import logging
import re
from contextlib import contextmanager
from pathlib import Path
from typing import Dict, Iterator, List, Optional

from filelock import FileLock
from git import Actor, GitError, Repo

from pygitops._util import checkout_pull_branch as _checkout_pull_branch
from pygitops._util import get_lockfile_path as _get_lockfile_path
from pygitops._util import is_git_repo as _is_git_repo
from pygitops._util import lock_repo as _lock_repo
from pygitops._util import push_error_present as _push_error_present
from pygitops._util import repo_working_dir as _repo_working_dir
from pygitops.exceptions import PyGitOpsError, PyGitOpsStagedItemsError
from pygitops.remote_git_utils import _scrub_github_auth
from pygitops.types import PathOrStr

_logger = logging.getLogger(__name__)


def stage_commit_push_changes(
    repo: Repo,
    branch_name: str,
    actor: Actor,
    commit_message: str,
    items_to_stage: Optional[List[Path]] = None,
    kwargs_to_push: Optional[Dict] = None,
) -> None:
    """
    Handles the logic of persisting filesystem changes to a local repository via a commit to a feature branch.

    This includes staging, committing, and pushing the local changes.

    :param repo: Repository object.
    :param branch_name: Feature branch from which changes will be committed and pushed.
    :param actor: The Github actor with which to perform the commit operation.
    :param commit_message: Text to be used as the commit message.
    :param items_to_stage: List of files and directories that will be staged for commit, will be inferred if parameter not provided.
        Please use an empty list to signal that there are intentionally no items to stage, and items_to_stage should not be inferred.
    :param kwargs_to_push: dictionary of arguments to pass to the push operation
    :raises PyGitOpsStagedItemsError: Items to stage are not present or could not be determined.
    :raises PyGitOpsError: There was an error staging, committing, or pushing code.
    """
    index = repo.index
    workdir_path = Path(_repo_working_dir(repo))

    # We will determine items_to_stage if the parameter was not provided.
    if items_to_stage is None:
        untracked_file_paths = [Path(f) for f in repo.untracked_files]
        items_to_stage = untracked_file_paths + [
            Path(f.a_path) for f in index.diff(None)
        ]

        if not items_to_stage:
            raise PyGitOpsStagedItemsError(
                "There are no items to stage, cannot perform commit operation"
            )

    # stage and commit changes using the provided actor.
    for item in items_to_stage:
        full_path = workdir_path / item
        index.add(str(item)) if full_path.exists() else index.remove(str(item), r=True)
    commit = index.commit(commit_message, author=actor, committer=actor)

    _logger.debug(
        f"Successfully made commit with stats: {commit.stats.files} to repository: {repo}"
    )

    # push changes to the remote branch
    origin = repo.remotes.origin
    if not kwargs_to_push:
        kwargs_to_push = {}
    push_info = origin.push(branch_name, **kwargs_to_push or {})[0]

    _logger.debug(
        f"Issued commit to remote branch: {branch_name}, with resulting summary: {push_info.summary} and flags: {push_info.flags}. (see flag documentation: https://gitpython.readthedocs.io/en/stable/reference.html#git.remote.PushInfo)"
    )

    if _push_error_present(push_info):
        raise PyGitOpsError(
            f"Unable to push to branch {branch_name} of new repo: {repo}"
        )


@contextmanager
def feature_branch(repo: Repo, branch_name: str) -> Iterator[None]:
    """
    Checkout the desired feature branch.

    Return to the default branch when context is exited.

    The main responsibility of this context manager is to limit concurrent
    operations on the local filesystem via the FileLock.

    Local git commit operations in particular should leverage this behavior.

    *** Assumes default branch is checked out, otherwise raises PyGitOpsError ***

    :param repo: Repository object
    :param branch_name: str object indicating the branch we would like to checkout
    :raises PyGitOpsError: There was an error performing the feature branch operation.
    """
    default_branch = get_default_branch(repo)

    untracked_files = repo.untracked_files
    if untracked_files:
        raise PyGitOpsError(
            f"We cannot checkout a feature branch when there are unstaged changes in your current branch: {untracked_files}"
        )

    active_branch = repo.active_branch
    if active_branch != repo.heads[default_branch]:
        raise PyGitOpsError(
            f"We can only checkout feature branches originating from the default branch: {default_branch}. Current branch is: {active_branch.name}"
        )

    _logger.debug(
        f"About to invoke `_lock_repo` for repo: {repo}, branch name: {branch_name}"
    )

    # lock the following operation such that the process of updating default branch,
    # checking out a feature branch, applying changes, and moving back to default branch never occurs concurrently
    with _lock_repo(repo):

        _logger.debug(
            f"Successfully using repository: {repo}, branch_name: {branch_name}"
        )

        # before creating a feature branch, checkout and update the default branch of the repository
        # helpful in the case where a failed process clears the FileLock, but leaves the repo on a feature branch
        origin = repo.remotes.origin
        # `origin.refs` might be out of date, this makes local checkout of repo aware of remote branches
        origin.fetch()

        # Handle the case where the remote is a bare repository with no commit history.
        # When there is no commit history, there is no default branch
        if default_branch in origin.refs:
            _checkout_pull_branch(repo, default_branch)
            _logger.debug(
                f"Successfully updated {default_branch} branch of repo: {repo}"
            )

        if branch_name != default_branch:
            # create and checkout a local feature branch
            feature_branch = repo.create_head(branch_name)
            feature_branch.checkout()

            _logger.debug(
                f"Successfully checked out feature branch: {branch_name} for repository: {repo}"
            )

        # give control back to call of this context manager
        try:
            yield
        finally:
            # clean up the feature branch
            repo.git.clean("-xdf")
            repo.git.reset("--hard")
            # move back to the repo's default branch when the `feature_branch` context is exited
            repo.heads[default_branch].checkout()
            _logger.debug(
                f"Successfully moved back to {default_branch} branch for repository: {repo} after using feature branch"
            )


def get_updated_repo(repo_url: str, clone_dir: PathOrStr, **kwargs) -> Repo:
    """
    Clone the default branch of the target repository, returning a repo object.

    If repo is already present, we will pull in the latest changes to the default branch of the repo.

    :param repo_url: URL of the Github repository to be cloned.
    :param clone_dir: The empty directory to clone repository content to.
    :raises PyGitOpsError: There was an error cloning the repository.
    """
    # make sure it's actually a Path if our user passed a str
    clone_dir = Path(clone_dir)

    # if clone dir does not exist, create it, and all parent dirs
    clone_dir.mkdir(parents=True, exist_ok=True)

    git_lockfile_path = _get_lockfile_path(str(clone_dir))

    # Lock the following operation such that only one process will attempt to clone the repo at a time.
    with FileLock(str(git_lockfile_path)):
        try:
            # if the repo already exists, don't clone it
            if _is_git_repo(clone_dir):
                repo = Repo(clone_dir)
                # pull down latest changes from `branch` if provided in kwargs, deferring to repo default branch
                branch = kwargs.get("branch") or get_default_branch(repo)
                # destroy any local changes to tracked and untracked files if `force` is provided in kwargs
                force = kwargs.get("force") or False
                _checkout_pull_branch(repo, branch, force=force)
                return repo

            # remove 'force' from kwargs if present, as it is not supported by clone
            kwargs.pop("force", None)
            return Repo.clone_from(repo_url, clone_dir, **kwargs)
        except GitError as e:
            clean_repo_url = _scrub_github_auth(repo_url)
            scrubbed_error_message = _scrub_github_auth(str(e))
            raise PyGitOpsError(
                f"Error cloning or updating repo {clean_repo_url} into destination path {clone_dir}: {scrubbed_error_message}"
            ) from e


def get_default_branch(repo: Repo) -> str:
    """
    Get the default branch of the provided repository.

    Although Git itself does not have a concept of default branches, a default branch on a GitHub repository maps directly to the HEAD symbolic ref.
    Pull in remote branches and update local head refs pointing to the remote repository.
    We expect the HEAD ref to match a particular pattern that we regex against.

    :param repo: git.Repo instance.
    :return: string representing name of default branch.
    """
    git_ref_regex = r"refs\/remotes\/origin\/([\w*\-\.]+)"
    symbolic_ref_head = "refs/remotes/origin/HEAD"

    # local repo should be aware of branch objects prior to running the `set-head` command, where an unknown branch might be present
    repo.remotes.origin.fetch()

    # update HEAD pointer before querying local state
    repo.git.remote(["set-head", "-a", "origin"])

    # query local state for the HEAD pointer
    default_ref = repo.git.symbolic_ref(symbolic_ref_head)

    match = re.match(git_ref_regex, default_ref)
    if not match:
        raise PyGitOpsError(
            f"None of the symbolic refs using {symbolic_ref_head} matched the regex: {git_ref_regex}"
        )

    expected_position = 1
    try:
        return match.group(expected_position)
    except IndexError:
        raise PyGitOpsError(
            f"The match object did not have a group in position {expected_position}"
        )



import shutil
import unittest.mock as mock
from pathlib import Path

import pytest
from filelock import Timeout
from git import PushInfo, Repo

from pygitops._util import (
    _lockfile_path,
    checkout_pull_branch,
    get_lockfile_path,
    is_git_repo,
    lock_repo,
    push_error_present,
    repo_working_dir,
)
from pygitops.exceptions import PyGitOpsError, PyGitOpsWorkingDirError

SOME_REPO_NAME = "some-repo-name"


@pytest.fixture(scope="session", autouse=True)
def _teardown():
    """This function removes the lockfiles directory after all tests are done."""
    yield  # let tests run
    shutil.rmtree(_lockfile_path)


@pytest.mark.parametrize(
    ("expected_response", "flags"),
    [(True, PushInfo.ERROR), (True, PushInfo.ERROR + 1), (False, PushInfo.ERROR - 1)],
)
def test_push_error_present(expected_response, flags):
    """
    Validate push_error_present reports an issue correctly.

    Given some integer representing the push flags, validate the reporting of each
    push error scenario.
    """

    push_info = PushInfo(flags, "", "", "")
    assert push_error_present(push_info) == expected_response


def test_lock_repo__filelock_acquire_timeout__raises_pygitops_error(mocker):
    class FakeFileLock:
        def acquire(self, timeout: int):
            raise Timeout("some-file")

    repo_mock = mocker.Mock(working_dir=f"some-repo-namespace/{SOME_REPO_NAME}")
    mocker.patch("pygitops._util.FileLock", return_value=FakeFileLock())

    with pytest.raises(PyGitOpsError):
        with lock_repo(repo_mock):
            pass


@pytest.mark.parametrize("path_exists", (True, False))
def test_get_lockfile_path__expected_path_returned(mocker, tmp_path, path_exists):

    mocker.patch("pygitops._util._lockfile_path", new=tmp_path)

    if path_exists:
        (tmp_path / SOME_REPO_NAME).mkdir()

    assert get_lockfile_path(SOME_REPO_NAME) == Path(
        f"{tmp_path}/{SOME_REPO_NAME}_lock_file.lock"
    )


def test_get_lockfile_path__lockfile_dir_dne__lockfile_dir_created(mocker, tmp_path):

    some_lockfile_dir = "some-lockfile-dir"
    mocker.patch("pygitops._util._lockfile_path", new=tmp_path / some_lockfile_dir)

    assert get_lockfile_path(SOME_REPO_NAME) == Path(
        f"{tmp_path}/{some_lockfile_dir}/{SOME_REPO_NAME}_lock_file.lock"
    )


def test_pull_branch__head_present__head_not_created(mocker):
    """
    In the case where checkout_pull_branch() is invoked, and a head for the specified branch is present,
    the operation should not create head nor set the tracking branch
    """

    branch = "test_branch"

    test_branch_head_mock = mocker.Mock()

    origin_mock = mocker.Mock(refs={branch: mock.ANY})
    create_head_mock = mocker.Mock()

    repo = mocker.Mock(
        heads={branch: test_branch_head_mock},
        remotes=mocker.Mock(origin=origin_mock),
        create_head=create_head_mock,
    )

    checkout_pull_branch(repo, branch)

    create_head_mock.assert_not_called()
    test_branch_head_mock.assert_not_called()


def test_pull_branch__head_not_present__head_created(mocker):
    """
    In the case where checkout_pull_branch() is invoked, and a head is not present,
    the operation should create the head and set the tracking branch
    """

    heads = {}
    branch = "test_branch"
    ref = mock.ANY

    test_branch_head_mock = mocker.Mock()

    def _create_head(*args):
        """
        Invoked as a side effect of the `create_head_mock`.
        Updates state of `heads` such that tracking branch registration succeeds
        """
        heads[branch] = test_branch_head_mock

    origin_mock = mocker.Mock(refs={branch: ref})
    create_head_mock = mocker.Mock(side_effect=_create_head)

    repo = mocker.Mock(
        heads=heads,
        remotes=mocker.Mock(origin=origin_mock),
        create_head=create_head_mock,
    )

    checkout_pull_branch(repo, branch)

    create_head_mock.assert_called_once_with(branch, ref)
    test_branch_head_mock.set_tracking_branch.assert_called_once()


def test_pull_branch__head_present_and_force_requested__tracked_and_untracked_changes_removed(
    mocker,
):
    """
    In the case where checkout_pull_branch() is invoked, and a head for the specified branch is present,
    and the "force" option is True, the operation should perform a force-checkout and remove untracked files.
    """
    branch = "test_branch"

    git_clean_mock = mocker.Mock()
    git_mock = mocker.Mock(clean=git_clean_mock)

    checkout_mock = mocker.Mock()

    test_branch_head_mock = mocker.Mock(
        checkout=checkout_mock,
    )

    origin_mock = mocker.Mock(refs={branch: mock.ANY})
    create_head_mock = mocker.Mock()

    repo = mocker.Mock(
        heads={branch: test_branch_head_mock},
        remotes=mocker.Mock(origin=origin_mock),
        create_head=create_head_mock,
        git=git_mock,
    )

    checkout_pull_branch(repo, branch, force=True)

    create_head_mock.assert_not_called()
    test_branch_head_mock.assert_not_called()
    checkout_mock.assert_called_once_with(force=True)
    git_clean_mock.assert_called_once()


def test_pull_branch__branch_not_in_refs__raises_pygitops_error(mocker):
    """
    In the case where the provided branch is not in the list of remote refs, A GitOperationError should be raised
    """

    origin_mock = mocker.Mock(refs={})

    repo = mocker.Mock(heads={}, remotes=mocker.Mock(origin=origin_mock))

    with pytest.raises(PyGitOpsError):
        checkout_pull_branch(repo, "test_branch")


def test_is_git_repo__not_a_git_repo__returns_false(tmp_path):
    assert is_git_repo(tmp_path) is False


def test_is_git_repo__is_a_git_repo__returns_true(tmp_path):
    Repo.init(tmp_path)
    assert is_git_repo(tmp_path)


def test_repo_working_dir__working_dir_none__raises_pygitops_working_dir_error(mocker):

    repo = mocker.Mock(working_dir=None)
    with pytest.raises(PyGitOpsWorkingDirError):
        repo_working_dir(repo)


def test_repo_working_dir__working_dir_present__expected_working_dir_returned(tmp_path):

    repo_path = tmp_path / SOME_REPO_NAME
    repo_path.mkdir()
    repo = Repo.init(repo_path)

    assert repo_working_dir(repo) == str(repo_path)



"""Utilities for working with git remotes."""

import re

from pygitops._constants import GITHUB_PUBLIC_DOMAIN_NAME


def build_github_repo_url(
    service_account_name: str,
    service_account_token: str,
    repo_namespace: str,
    repo_name: str,
    domain_name: str = GITHUB_PUBLIC_DOMAIN_NAME,
) -> str:
    """
    Build github repo_url with the intention of providing as the `repo_url` argument to `get_updated_repo`.
    This factory function is specific to github, and might not support url schemas of different git services.

    :param service_account_name: Name of the Github service account used to clone repository.
    :param service_account_token: Raw token associated with the provided Github service account name.
    :param repo_namespace: The namespace of the target repository.
    :param repo_name: The name of the target repository.
    :param domain_name: Github domain, allows us to point to instances other than public Github.
    """

    return f"https://{service_account_name}:{service_account_token}@{domain_name}/{repo_namespace}/{repo_name}.git"


def _scrub_github_auth(string: str) -> str:
    sa_token_regex = "https://.+?:.+?@"  # nosec
    sa_token_replace_term = "https://***:***@"  # nosec

    return re.sub(sa_token_regex, sa_token_replace_term, string)



class PyGitOpsError(Exception):
    """Parent exception class for all errors coming from `pygitops` package."""


class PyGitOpsValueError(PyGitOpsError):
    """The provided argument is invalid."""


class PyGitOpsStagedItemsError(PyGitOpsError):
    """There were no items to stage for commit."""


class PyGitOpsWorkingDirError(PyGitOpsError):
    """There was an error with the filesystem, namely `git.Repo.working_dir` is unexpectedly None."""



import pytest

from pygitops._constants import GITHUB_PUBLIC_DOMAIN_NAME
from pygitops.remote_git_utils import _scrub_github_auth, build_github_repo_url
from tests.test_operations import (
    SOME_GITHUB_DOMAIN_NAME,
    SOME_REPO_NAME,
    SOME_REPO_NAMESPACE,
    SOME_SERVICE_ACCOUNT_NAME,
    SOME_SERVICE_ACCOUNT_TOKEN,
)


def test_build_github_repo_url__expected_repo_url_returned():
    assert (
        build_github_repo_url(
            SOME_SERVICE_ACCOUNT_NAME,
            SOME_SERVICE_ACCOUNT_TOKEN,
            SOME_REPO_NAMESPACE,
            SOME_REPO_NAME,
            SOME_GITHUB_DOMAIN_NAME,
        )
        == f"https://{SOME_SERVICE_ACCOUNT_NAME}:{SOME_SERVICE_ACCOUNT_TOKEN}@{SOME_GITHUB_DOMAIN_NAME}/{SOME_REPO_NAMESPACE}/{SOME_REPO_NAME}.git"
    )


def test_build_github_repo_url__github_domain_name_not_provided__public_github_url_built():
    assert (
        build_github_repo_url(
            SOME_SERVICE_ACCOUNT_NAME,
            SOME_SERVICE_ACCOUNT_TOKEN,
            SOME_REPO_NAMESPACE,
            SOME_REPO_NAME,
        )
        == f"https://{SOME_SERVICE_ACCOUNT_NAME}:{SOME_SERVICE_ACCOUNT_TOKEN}@{GITHUB_PUBLIC_DOMAIN_NAME}/{SOME_REPO_NAMESPACE}/{SOME_REPO_NAME}.git"
    )


@pytest.mark.parametrize(
    ("input_", "output"),
    [
        (
            "https://some-service-account-name:test_cred_123@some-github-domain-name/test_namespace/some-repo-name.git",
            "https://***:***@some-github-domain-name/test_namespace/some-repo-name.git",
        )
    ],
)
def test__scrub_github_auth_1(input_, output):
    """
    Validate Github auth tokens are scrubbed from URLs
    """
    assert _scrub_github_auth(input_) == output
